#!/usr/bin/env node
'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var puppeteer = _interopDefault(require('puppeteer'));
var url = require('url');
var chalk = _interopDefault(require('chalk'));
var prettyBytes = _interopDefault(require('pretty-bytes'));
var columnify = _interopDefault(require('columnify'));
var yargs = _interopDefault(require('yargs'));

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

async function setPageOptions(page, options) {
  let { cookies, headers, userAgent, viewport } = options;

  if (typeof cookies !== 'undefined') {
    if (!Array.isArray(cookies)) {
      cookies = [cookies];
    }
    await page.setCookie(...cookies);
  }

  if (typeof headers !== 'undefined') {
    await page.setExtraHTTPHeaders(headers);
  }

  if (typeof userAgent !== 'undefined') {
    await page.setUserAgent(userAgent);
  }

  if (typeof viewport !== 'undefined') {
    await page.setViewport(viewport);
  }
}

var runner = (async function (urlString, options = {}) {
  const url$$1 = url.parse(urlString || '');

  if (!url$$1 || !url$$1.protocol) {
    throw new Error('A valid url is required.');
  }

  const { args = [], timeout } = options,
        browser = await puppeteer.launch({ args: Array.isArray(args) ? args : [args], timeout }),
        page = await browser.newPage(),
        { coverage } = page;

  // Set default options
  options = _extends({ js: true, css: true }, options);
  await setPageOptions(page, options);

  // Map the requests to match covered requests later
  let requestMap = new Map();
  page.on('request', request => {
    let requestUrl = request.url(),
        req = url.parse(requestUrl);
    // console.log({requestUrl})
    requestMap.set(requestUrl, _extends({
      type: request.resourceType(),
      headers: request.headers(),
      sameOrigin: !req.host ? true : req.host === url$$1.host
    }, req));
  });

  const pageFinished = new Promise(resolve => {
    page.on('load', resolve);
  });

  await Promise.all([options.js && coverage.startJSCoverage(), options.css && coverage.startCSSCoverage()]);

  try {
    await page.goto(urlString);
  } catch (ex) {
    await browser.close();
    throw ex;
  }

  await pageFinished;

  const [jsCoverage, cssCoverage] = await Promise.all([options.js ? coverage.stopJSCoverage() : Promise.resolve([]), options.css ? coverage.stopCSSCoverage() : Promise.resolve([])]);

  let assets = [],
      entries = [...jsCoverage, ...cssCoverage];

  if (options.sameOrigin) {
    entries = entries.filter(entry => requestMap.get(entry.url).sameOrigin);
  }

  for (const entry of entries) {
    let totalBytes = entry.text.length,
        usedBytes = 0,
        req = requestMap.get(entry.url) || {};

    for (const range of entry.ranges) {
      usedBytes += range.end - range.start;
    }

    assets.push({
      path: req.sameOrigin ? req.path : req.href,
      type: req.type,
      totalBytes,
      usedBytes,
      unusedBytes: totalBytes - usedBytes,
      coverage: totalBytes !== usedBytes ? usedBytes / totalBytes * 100 : 100
    });
  }

  await browser.close();

  let totalBytes = assets.reduce((totalBytes, assets) => totalBytes + assets.totalBytes, 0),
      totalUsedBytes = assets.reduce((totalUsedBytes, assets) => totalUsedBytes + assets.usedBytes, 0);

  return {
    url: urlString,
    host: url.parse(url$$1).host,
    totalBytes,
    totalUsedBytes,
    totalCoverage: totalUsedBytes / totalBytes * 100,
    coverages: assets
  };
});

function colorUsed(usedBytes, totalBytes) {
  let used = prettyBytes(usedBytes);

  if (totalBytes && usedBytes === 0) {
    used = chalk.red(used);
  } else {
    used = chalk.gray(used);
  }

  return used;
}

function colorUnused(unusedBytes, totalBytes) {
  let unused = prettyBytes(unusedBytes);

  if (unusedBytes && unusedBytes === totalBytes) {
    unused = chalk.red(unused);
  } else if (unusedBytes > 0) {
    unused = chalk.yellow(unused);
  } else {
    unused = chalk.green(unused);
  }

  return unused;
}

function colorCoverage(coveragePercent) {
  let precision = parseFloat(coveragePercent).toPrecision(coveragePercent < 10 ? 3 : 4),
      coverage = `${~~coveragePercent !== coveragePercent ? precision : coveragePercent}%`;

  if (coveragePercent === 100) {
    coverage = chalk.green(coverage);
  } else if (coveragePercent > 0) {
    coverage = chalk.yellow(coverage);
  } else {
    coverage = chalk.red(coverage);
  }

  return coverage;
}

function stats (coverage) {
  let { host, coverages } = coverage;

  coverages = coverages.map(entry => {
    let { path, type, totalBytes, usedBytes, unusedBytes, coverage: coveredPercent } = entry;

    const total = prettyBytes(totalBytes);
    const used = colorUsed(usedBytes, totalBytes);
    const unused = colorUnused(unusedBytes, totalBytes);
    const coverage = colorCoverage(coveredPercent);

    return { path, type, total, used, unused, coverage };
  });

  // Set right column alignments for numeric values
  const config = ['total', 'used', 'unused', 'coverage'].reduce((o, name) => (o[name] = { align: 'right' }) && o, {});

  const columns = columnify(coverages, { truncate: true, maxWidth: 120, minWidth: 8, config });

  return `${host} Coverage\n\n${columns}`;
}

const BASIC_OPTIONS = 'Basic options:';
const PUPPETEER_OPTIONS = 'Puppeteer options:';

const _yargs = yargs.usage('\nUsage: $0 <url> [options]').options({

  // Basic Options
  'no-css': {
    type: 'boolean',
    describe: 'disables css code coverage',
    group: BASIC_OPTIONS
  },
  'no-js': {
    type: 'boolean',
    describe: 'disables js code coverage',
    group: BASIC_OPTIONS
  },
  origin: {
    type: 'boolean',
    alias: 'o',
    describe: 'only cover same-origin assets',
    group: BASIC_OPTIONS
  },

  // Puppeteer Options
  timeout: {
    type: 'number',
    alias: 't',
    describe: 'maximum number of milliseconds to wait for the browser to start',
    group: PUPPETEER_OPTIONS
  },
  viewportWidth: {
    type: 'number',
    alias: 'vw',
    describe: 'sets viewport width for page render',
    group: PUPPETEER_OPTIONS
  },
  viewportHeight: {
    type: 'number',
    alias: 'vh',
    describe: 'sets viewport height for page render',
    group: PUPPETEER_OPTIONS
  },
  viewport: {
    type: 'string',
    alias: 'V',
    describe: 'sets viewport for page render i.e. 800x600',
    group: PUPPETEER_OPTIONS
  },
  userAgent: {
    type: 'string',
    alias: 'ua',
    describe: 'sets the user agent for the request',
    group: PUPPETEER_OPTIONS
  },
  headers: {
    type: 'string',
    alias: 'H',
    describe: 'sets headers for the page request\nexample: $0 <url> -H "X-Header: 123"',
    group: PUPPETEER_OPTIONS
  }

}).array('headers').alias('help', 'h').alias('version', 'v').help();

function parseHeaders(headers) {
  return headers.reduce((headers, headerString) => {
    let [key, value] = headerString.split(/:\s+/);

    if (key && value) {
      headers[key] = value;
    }

    return headers;
  }, {});
}

(async function () {
  let [url$$1] = _yargs.argv._,
      { origin, noCss, noJs, viewportHeight, viewportWidth, viewport, headers, userAgent } = yargs.argv;

  if (!url$$1) {
    yargs.showHelp();
    return;
  }

  if (viewport) {
    [viewportWidth, viewportHeight] = viewport.split('x').map(x => parseInt(x, 10) || void 0);
  }

  if (typeof viewportWidth === 'undefined' && viewportHeight && typeof viewportHeight === 'number') {
    viewportWidth = ~~(viewportHeight * (4 / 3));
  }

  if (typeof viewportHeight === 'undefined' && viewportWidth && typeof viewportWidth === 'number') {
    viewportHeight = ~~(viewportHeight * (3 / 4));
  }

  if (viewportHeight && viewportWidth) {
    viewport = { height: viewportHeight, width: viewportWidth };
  }

  if (headers) {
    headers = parseHeaders(headers);
  }

  try {
    const coverage = await runner(url$$1, { sameOrigin: origin, css: !noCss, js: !noJs, viewport, headers, userAgent });
    console.log(stats(coverage));
  } catch (err) {
    console.log(`${chalk.red('Error:')} ${err.message}`);
  }
})();
