'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var puppeteer = _interopDefault(require('puppeteer'));
var url = require('url');

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

async function setPageOptions(page, options) {
  let { cookies, headers, userAgent, viewport } = options;

  if (typeof cookies !== 'undefined') {
    if (!Array.isArray(cookies)) {
      cookies = [cookies];
    }
    await page.setCookie(...cookies);
  }

  if (typeof headers !== 'undefined') {
    await page.setExtraHTTPHeaders(headers);
  }

  if (typeof userAgent !== 'undefined') {
    await page.setUserAgent(userAgent);
  }

  if (typeof viewport !== 'undefined') {
    await page.setViewport(viewport);
  }
}

var runner = (async function (urlString, options = {}) {
  const url$$1 = url.parse(urlString || '');

  if (!url$$1 || !url$$1.protocol) {
    throw new Error('A valid url is required.');
  }

  const { args = [], timeout } = options,
        browser = await puppeteer.launch({ args: Array.isArray(args) ? args : [args], timeout }),
        page = await browser.newPage(),
        { coverage } = page;

  // Set default options
  options = _extends({ js: true, css: true }, options);
  await setPageOptions(page, options);

  // Map the requests to match covered requests later
  let requestMap = new Map();
  page.on('request', request => {
    let requestUrl = request.url(),
        req = url.parse(requestUrl);
    // console.log({requestUrl})
    requestMap.set(requestUrl, _extends({
      type: request.resourceType(),
      headers: request.headers(),
      sameOrigin: !req.host ? true : req.host === url$$1.host
    }, req));
  });

  const pageFinished = new Promise(resolve => {
    page.on('load', resolve);
  });

  await Promise.all([options.js && coverage.startJSCoverage(), options.css && coverage.startCSSCoverage()]);

  try {
    await page.goto(urlString);
  } catch (ex) {
    await browser.close();
    throw ex;
  }

  await pageFinished;

  const [jsCoverage, cssCoverage] = await Promise.all([options.js ? coverage.stopJSCoverage() : Promise.resolve([]), options.css ? coverage.stopCSSCoverage() : Promise.resolve([])]);

  let assets = [],
      entries = [...jsCoverage, ...cssCoverage];

  if (options.sameOrigin) {
    entries = entries.filter(entry => requestMap.get(entry.url).sameOrigin);
  }

  for (const entry of entries) {
    let totalBytes = entry.text.length,
        usedBytes = 0,
        req = requestMap.get(entry.url) || {};

    for (const range of entry.ranges) {
      usedBytes += range.end - range.start;
    }

    assets.push({
      path: req.sameOrigin ? req.path : req.href,
      type: req.type,
      totalBytes,
      usedBytes,
      unusedBytes: totalBytes - usedBytes,
      coverage: totalBytes !== usedBytes ? usedBytes / totalBytes * 100 : 100
    });
  }

  await browser.close();

  let totalBytes = assets.reduce((totalBytes, assets) => totalBytes + assets.totalBytes, 0),
      totalUsedBytes = assets.reduce((totalUsedBytes, assets) => totalUsedBytes + assets.usedBytes, 0);

  return {
    url: urlString,
    host: url.parse(url$$1).host,
    totalBytes,
    totalUsedBytes,
    totalCoverage: totalUsedBytes / totalBytes * 100,
    coverages: assets
  };
});

async function quilted(url$$1, options) {
  return await runner(url$$1, options);
}

module.exports = quilted;
